;; Programa: Package con agentes jugadores simples de PlanTron
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    12/04/2006
;; Comentarios:
;;
;;           Incluye agentes para jugar al PlanTron simples, esto
;;           es, agentes que no tiene memoria y su comportamiento
;;           viene guiado en función del entorno que perciben en
;;           cada instante (excepto el agente-hormiga-borracha,
;;           que hace uso del generador de números aleatorios).
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT-BASE y ENVIO.
;;
;;           Los agentes incluidos son, ordenados según la
;;           complejidad de su código, los siguientes:
;;
;;             AGENTE-MITAD
;;             AGENTE-TEMERARIO
;;             AGENTE-ESPIRAL
;;             AGENTE-IMITADOR
;;             AGENTE-HORMIGA-BORRACHA
;;
;;           Ver la documentación existente para cada una de ellos.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (13/04/2006): Primera versión. Extraida
;;             de tron.0.1.lsp. Vease tron-base.lsp y tron-comm*.lsp
;;
;;           - Ver 0.2 (21/04/2006): Revisados nombre de variables y
;;             funciones. Transformado todo a funciones recursivas.
;;             Algunas simplificaciones del código. Ampliada
;;             documentación.
;;
;;           - Ver 0.3 (29/04/2006): Empaquetado, puesto interface
;;             de definición de agentes (const agentes). Ofuscado
;;             código de dir-vacia-pref.
;;
;;           - Ver 0.3.1 (02/04/2006): Incluido agente mitad.
;;
;;           - Ver 0.3.2 (05/04/2006): Cambiado nombre constante
;;             *version-tron-agentes-simples*.
;;
;;           - Ver 0.4 (21/4/2007): Cambio de "Tron" a "Plantron":
;;             Cambio de nombre de "tron-agentes-simples" a 
;;             "pt-agentes-simples". Cambio extensión de ".lsp" a
;;             "lisp". Cambio también nombre de variables,
;;             funciones, etc.
;;
;;           - Ver 0.4.1 (27/4/2007): Revisión menor del código y 
;;             documentación.
;;
;;           - Ver 0.5 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;; Por hacer:
;;
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-AGENTES-SIMPLES"
  (:nicknames "PT-A-S")
  (:use "COMMON-LISP")
  (:export "AGENTES" "VERSION" "AGENTE-MITAD" "AGENTE-TEMERARIO" "AGENTE-ESPIRAL" "AGENTE-HORMIGA-BORRACHA" "AGENTE-IMITADOR" "DIR-POSIBLE-P" "DIR-INC" "DIR-POSIBLE-PREF")
  (:documentation "Agentes para jugar al PlanTron simples, esto es, agentes que no tiene memoria y su comportamiento viene guiado en función del entorno que perciben en cada instante (excepto el agente-hormiga-borracha, que hace uso del generador de números aleatorios)."))


(in-package "PT-AGENTES-SIMPLES")


;;
;; Constante con lista de agentes presentes en este package
;; --------------------------------------------------------

(defconstant agentes
  '((:turno agente-mitad)
    (:turno agente-temerario)
    (:turno agente-espiral)
    (:turno agente-imitador)
    (:turno agente-hormiga-borracha))
  "Agentes simples para jugar a PlanTron.")


;;
;; Constante versión
;; -----------------

(defconstant version "0.5"
  "Versión del package PT-AGENTES-SIMPLES.")


;;
;; Funciones de tratamiento de direcciones y posiciones
;; ----------------------------------------------------

(defun libre-o-planta (tipo)
  "Devuelve T si TIPO (un valor devuelto por ENVIO:CONSULTA-TIPO) se corresponde a una posición vacía (NIL) o a una planta (ENVIO:PLANTA)"
  (or (not tipo) (eq tipo 'envio:planta)))

(defun dir-posible-p (pos dir)
  "Mira si es posible colocarse (sin obstáculo o rastro de agente con que colisionar) en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve DIR en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente)."
  (if (libre-o-planta (envio:consulta-tipo (envio:consulta-pos-tab pos dir)))
      dir))

(defun number-o-0 (elm)
  "Si ELM es un número lo devuelve, sino devuelve 0."
  (if (numberp elm)
      elm
    0))

(defun dir-inc (dir dir-ini)
  "Calcula la dirección DIR relativa a DIR-INI, esto es, gira DIR-INI pasos en sentido agujas del reloj sobre DIR. DIR y DIR-INI han de ser números enteros (direcciones,  numero entero entre 0 y 3: norte, este, sur y oeste respectivamente), si no se consideran cero (norte)."
  (rem (+ (number-o-0 dir) (number-o-0 dir-ini)) 4))

(defun dir-posible-pref (pos lis-dir-pref &optional dir-ini)
  "Busca la primera dirección posible (ver función DIR-POSIBLE-P) alrededor de POS siguiendo el orden de LIS-DIR-PREF, si DIR-INI está instanciado las direcciones de LIS-DIR-PREF son relativas a DIR-INI. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. LIS-DIR-PREF ha de ser una lista con números enteros (direcciones). DIR-INI ha de ser un número entero (dirección)."
  (if (consp lis-dir-pref)
      (or (dir-posible-p pos (dir-inc (first lis-dir-pref) dir-ini))
	  (dir-posible-pref pos (rest lis-dir-pref) dir-ini))))


;;
;; Agente NSEO
;; -----------

(defun agente-mitad (id)
  "Agente jugador PlanTron que sólo busca la mitad."
  (dir-posible-pref (envio:consulta-pos-jug id) '(0 2 1 3)))


;;
;; Agente rombo derecha
;; --------------------

(defun agente-temerario (id)
  "Agente jugador PlanTron que va temerariamente desenrrollándose!"
  (dir-posible-pref (envio:consulta-pos-jug id)
		    '(1 3 0)
		    (first (envio:consulta-est-jug id))))


;;
;; Agente espiral derecha
;; ----------------------
  
(defun agente-espiral (id)
  "Agente jugador PlanTron que da vueltas en espiral desenrrollándose!"
  (dir-posible-pref (envio:consulta-pos-jug id)
		    '(1 0 3)
		    (first (envio:consulta-est-jug id))))




;;
;; Agente imitador
;; ---------------

(defun agente-imitador (id)
  "Agente jugador PlanTron que trata de hacer el mismo movimiento que su primer adversario, si tiene."
  (dir-posible-pref (envio:consulta-pos-jug id) 
		  '(0 1 3 2)
		  ; Extraemos dirección del primer adversario
		  (first (envio:consulta-est-jug (if (eq id 1) 2 1)))))



;;
;; Agente hormiga borracha
;; -----------------------

(defvar random-state-hormiga-borracha (make-random-state *random-state*)
  "Estado random para el AGENTE-HORMIGA-BORRACHA.")

(defun agente-hormiga-borracha (id)
  "Agente jugador PlanTron que da tumbos y se agarra a las farolas. Elige una dirección aleatorias entre las vacías a su alrededor."
  (let* ((pos (envio:consulta-pos-jug id))
	 (lis-dir (remove nil (mapcar (lambda (dir) (dir-posible-p pos dir))
				       '(0 1 2 3)))))
    (if lis-dir
	(nth (random (length lis-dir) random-state-hormiga-borracha)
	     lis-dir))))

